/* This file is part of the Roméo model-checking software

Copyright École Centrale de Nantes, LS2N

Contributors: Didier Lime (2014-2024)

Didier.Lime@ec-nantes.fr

This software is a computer program whose purpose is to perform
parametric model checking on timed and hybrid systems.

This software is governed by the CeCILL license under French law and
abiding by the rules of distribution of free software.  You can  use, 
modify and/ or redistribute the software under the terms of the CeCILL
license as circulated by CEA, CNRS and INRIA at the following URL
"http://www.cecill.info". 

As a counterpart to the access to the source code and  rights to copy,
modify and redistribute granted by the license, users are provided only
with a limited warranty  and the software's author,  the holder of the
economic rights,  and the successive licensors  have only  limited
liability. 

In this respect, the user's attention is drawn to the risks associated
with loading,  using,  modifying and/or developing or reproducing the
software by the user in light of its specific status of free software,
that may mean  that it is complicated to manipulate,  and  that  also
therefore means  that it is reserved for developers  and  experienced
professionals having in-depth computer knowledge. Users are therefore
encouraged to load and test the software's suitability as regards their
requirements in conditions enabling the security of their systems and/or 
data to be ensured and,  more generally, to use and operate it in the 
same conditions as regards security. 

The fact that you are presently reading this means that you have had
knowledge of the CeCILL license and that you accept its terms. */

#include <control_result.hh>
#include <cts.hh>
#include <transition.hh>
#include <color_output.hh>

using namespace romeo;
using namespace std;

#include <logger.hh> 
extern romeo::Logger Log;

ControlResult::ControlResult(PResult* ir, bool r, wset_type t): reach(r), internal_result(ir), wstype(t)
{
    if (!reach)
    {
        ir->negation();
    }
}

ControlResult::ControlResult(PResult* ir, bool r, const PassedGN& p): reach(r), internal_result(ir), wstype(p.wstype)
{
    if (!reach)
    {
        ir->negation();
    }

    strategy = p.get_strategy();
}

PResult* ControlResult::copy() const
{
    return new ControlResult(internal_result->copy(), reach, wstype);
}

Polyhedron ControlResult::constraint() const
{
    return internal_result->constraint();
}

PResult& ControlResult::conjunction(const PResult& r)
{
    internal_result->conjunction(r);
    return *this;
}

PResult& ControlResult::disjunction(const PResult& r)
{
    internal_result->disjunction(r);
    return *this;
}

PResult& ControlResult::negation()
{
    internal_result->negation();
    return *this;
}

bool ControlResult::empty() const
{
    return internal_result->empty();
}

bool ControlResult::universe() const
{
    return internal_result->universe();
}

bool ControlResult::equals(const PResult&)
{
    cerr << error_tag(color_output) << "Cannot test ControlResults for equality" << endl;
    exit(1);
}

bool ControlResult::contains(const PResult&)
{
    cerr << error_tag(color_output) << "Cannot test ControlResults for inclusion" << endl;
    exit(1);
}

bool ControlResult::intersects(const PResult&)
{
    cerr << error_tag(color_output) << "Cannot test ControlResults for intersection" << endl;
    exit(1);
}


string ControlResult::to_string() const
{
    return internal_result->to_string();
}

string ControlResult::get_trace() const
{    
    stringstream s;
 
    //s << endl;
    if (strategy.empty())
    {
        s << endl << "No (partial) strategy for the ";
        if (reach)
        {
            s << "controller." << endl;
        } else {
            s << "environment." << endl;
        }
    } else {
        if (reach)
        {
            if (!internal_result->empty())
            {
                s << endl << "Strategy ";
            } else {
                s << endl << "Partial strategy ";
            }
            s << "for the controller:" << endl;
        } else {
            if (!internal_result->empty())
            {
                s << endl << "Partial strategy ";
            } else {
                s << endl << "Strategy ";
            }
            s << "for the environment:" << endl;
        }
 
        for (auto si: strategy)
        {
            // Discrete part
            s << "when " << cts->variables.value_to_string(si.first) << endl;

            // Action part
            if (si.second.empty())
            {
                s << "  wait";
            } else {
                for (auto sj: si.second)
                {
                    const Transition* t = sj.action;
                    if (wstype == WSET_UNTIMED)
                    {
                        if (t != NULL)
                        {
                            s << "  do " << t->label;
                            if (!sj.domain.wb)
                            {
                                s << " immediately";
                            }
                        } else {
                            s << "  wait";
                        }
                    } else {
                        string domain;
                        vector<string> labels;
                        if (wstype == WSET_POLY)
                        {
                            for (list<Parameter>::const_iterator k=cts->begin_parameters(); k != cts->end_parameters(); k++)
                            {
                                cvalue v;
                                if (!k->constant(v))
                                    labels.push_back(k->label);
                            }
    
                            const unsigned NP = cts->nparams();
     
                            for (auto tk: sj.en)
                            {
                                labels.push_back(tk->label);
                            }

                            if (cts->has_cost())
                            {
                                labels.push_back("#cost");
                            }
    
                            domain = sj.domain.wp->to_string_labels(labels, NP);
    
                        } else if (wstype == WSET_DBM) {
                            labels.push_back("0");
                            for (auto tk: sj.en)
                            {
                                labels.push_back(tk->label);
                            }
    
                            domain = sj.domain.wd->to_string_labels(labels);
                        }
    
                        if (sj.type == STRAT_ACTION)
                        {
                            s << "with: " << endl << domain << endl; 
                            if (t != nullptr)
                            {
                                s << "   do " << t->label;
                            } else {
                                s << "   wait";
                            }
                            s << endl << endl;
                        } else if (sj.type == STRAT_ACTION_CLASS) {
                            s << "if initialisation results in: " << endl << domain << endl; 
                            if (t != nullptr)
                            {
                                s << "   (wait if necessary and) do " << t->label;
                            } else {
                                s << "   wait"; // should not be possible
                            }
                            s << endl << endl;

                        } else {
                            if (!sj.ctr.empty())
                            {
                                if (t == nullptr)
                                {
                                    s << "initialize ";
                                } else {
                                    s << "arriving with " << t->label << ", initialize ";
                                }
                                bool comma = false;
                                for (auto tk: sj.ctr)
                                {
                                    if (comma)
                                    {
                                        s << ", ";
                                    } else {
                                        comma = true;
                                    }
                                    s << tk->label;
                                }

                                s << " such that:" << endl << domain << endl;
                            } 
                            //else {
                            //    s << "  wait" << endl;
                            //}
                        }
                    }
                }
            }
            s << endl;
        }
    }
 
    return colorize(color_output, s.str(), AC_CYAN);
}

